#!/usr/bin/env python

"""
Perform deductions on an inspected program.

Copyright (C) 2006, 2007, 2010, 2011, 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.errors import *
import compiler.ast

# Source code classes.

class DeducedSource(ASTVisitor):

    "A module upon which deductions of code behaviour are made."

    def __init__(self, module, program):
        self.visitor = self
        self.module = module
        self.program = program
        self.objtable = program.get_object_table()
        self.units = []
        self.expr = None

    def get_unit(self):
        return self.units[-1]

    def get_module(self):
        return self.units[0]

    def deduce(self):

        "Process the module, making deductions."

        self.dispatch(self.module.astnode)

    def dispatch(self, node, *args):

        "NOTE: From compiler.visitor.ASTVisitor."

        try:
            return node.visit(self.visitor, *args)
        except AttributeError:
            # NOTE: Obligatory hack to find real attribute errors.
            if isinstance(node, (Getattr, AssAttr)):
                raise
            return self.visitor.default(node, *args)

    def _visitUnit(self, node):

        """
        Track entry into program units in order to support various attribute
        access operations.
        """

        self.units.append(node.unit)
        self.dispatch(node.node)
        self.units.pop()

    def _visitOptionalUnit(self, node):

        "Optionally visit a unit, depending on whether it is used."

        if not used_by_unit(node):
            return
        self._visitUnit(node)

    visitModule = _visitUnit
    visitClass = visitFunction = _visitOptionalUnit

    def _visitAttr(self, node):

        """
        Perform deductions on attribute accesses, adding annotations to the node
        that can be used by subsequent activities.
        """

        unit = self.get_unit()

        # Remember to permit deductions on the expression node. Here, we may
        # also obtain a concrete type associated with an instantiation.

        expr_type = self.dispatch(node.expr)
        if not node._expr or isinstance(node._expr, Instance):
            node._expr = expr_type

        # The target, on which the access is performed, may influence the effect
        # on the context. We can only reliably assume that a literal constant is
        # an instance: all other "instances" may actually be classes in certain
        # cases.

        target = node._expr
        instance_target = isinstance(target, TypedInstance)
        typed_instance_attr = isinstance(target, Attr) and isinstance(target.get_value(), TypedInstance)
        self_access = self.provides_self_access(node, unit)

        # Attempt to deduce attributes from explicit annotations.

        node._attrs_deduced = attrs = self.possible_attributes_from_annotation(node)

        if len(attrs) == 1:
            for attr, value in attrs:

                # Constant values can be obtained directly.

                if self.provides_constant_result(value):
                    node._access_type = "constant"
                    node._value_deduced = value
                    return

                # Static attributes can be obtained via their parent.

                if attr.is_static_attribute():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = instance_target and "set" or None
                    return

        # If a reliable target was originally specified, any usable attributes
        # should have been detected above, and any attributes deduced by other
        # means will not be compatible with the target. Thus, the nature of the
        # target is investigated: it must be an inspectable namespace or be an
        # attribute only providing such namespaces; otherwise, it is possible
        # that deduced attributes might be appropriate.

        if target and (
            isinstance(target, (Class, Module)) or
            isinstance(target, Attr) and not [v for v in target.get_values() if not isinstance(v, (Class, Module))]
            ):
            node._access_type = "impossible"
            return

        # Attributes of self, which is by definition an instance, or typed
        # instances, which act somewhat like self in that their properties
        # should be known.

        if instance_target or typed_instance_attr or self_access:

            if instance_target:
                value = target
            elif typed_instance_attr:
                value = target.get_value()

            # Find the class of the instance.

            if instance_target or typed_instance_attr:
                if isinstance(value, Const):
                    cls = get_constant_class(value.get_class_name())
                else:
                    cls = value.cls
            else:
                cls = unit.parent

            # Find instance attributes.

            attr = cls.instance_attributes().get(node.attrname)

            # Where self is involved, descendants can also provide attributes.

            attrs = self_access and filter(None, [desc.instance_attributes().get(node.attrname) for desc in cls.descendants]) or []

            # A "leaf" class whose instances provide an attribute.

            if attr and not attrs:
                node._access_type = "instance"
                node._attr_deduced = attr
                return

            # A class where instances of subclasses may provide an attribute.

            elif attrs:
                if attr:
                    attrs.append(attr)

                node._attrs_deduced = [(a, a.get_value()) for a in attrs]

                # The instances may provide the attribute at the same position.

                positions = set([a.position for a in attrs])
                if len(positions) == 1:
                    for position in positions:
                        node._access_type = "positioned"
                        node._position_deduced = position
                        return

                # Otherwise, accessing the attributes is more work.

                node._access_type = "instance"
                return

            # Find class attributes.
            # The context will be overridden for compatible class attributes
            # only.

            attr = cls.all_class_attributes().get(node.attrname)

            if attr:

                # Constant attributes.

                if attr.is_strict_constant():
                    if self.provides_constant_result(attr.get_value()):
                        node._access_type = "constant"
                        node._value_deduced = attr.get_value()
                        node._set_context = "set"
                        return

                # Compatible class attributes.

                if attr.defined_within_hierarchy():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = "set"
                    return

                # Incompatible class attributes.

                elif attr.defined_outside_hierarchy():
                    node._access_type = "static"
                    node._attr_deduced = attr
                    return

                # Unknown or mixed compatibility.

                node._access_type = "static"
                node._attr_deduced = attr
                node._set_context = "cond"
                return

        # Usage observations, both specific to this node's region of the program
        # and also applicable to the lifespan of the affected name.

        specific_targets = self.possible_accessors_from_usage(node, defining_users=0)
        targets = self.possible_accessors_from_usage(node, defining_users=1)

        # Record whether types were already deduced. If not, get types using
        # only this attribute.

        if not specific_targets or not targets:
            attribute_targets = self.possible_accessors_for_attribute(node.attrname)
            if not specific_targets:
                specific_targets = attribute_targets
            if not targets:
                targets = attribute_targets

        # Get the attributes from the deduced targets.

        node._attrs_deduced_from_specific_usage = self.get_attributes(specific_targets, node.attrname)
        node._attrs_deduced_from_usage = attrs = self.get_attributes(targets, node.attrname)

        # Generate optimisations where only a single attribute applies.

        if attrs and len(attrs) == 1:
            for attr in attrs:

                # Static attributes, but potentially non-static targets.

                if attr.is_static_attribute():

                    # Static attributes may be accompanied by a different context
                    # depending on the accessor.
                    # NOTE: Should determine whether the context is always replaced.

                    node._access_type = "static"
                    node._attr_deduced = attr
                    node._set_context = instance_target and "set" or "cond"
                    return

                # Non-static attributes.

                node._access_type = "instance"
                node._attr_deduced = attr
                return

        # Test for compatible attribute positioning.

        elif attrs:
            positions = set([(attr.is_static_attribute(), attr.position) for attr in attrs])

            # Permit a position-based access only on non-static attributes since
            # access to static attributes may happen via instances and thus not
            # be relative to the accessor but to its parent.

            if len(positions) == 1:
                for position in positions:
                    if not position[0]:
                        node._access_type = "positioned"
                        node._position_deduced = position[0]
                        return

        # With no usable deductions, generate a table-based access.

        node._access_type = "unknown"
        node._set_context = "cond"

    visitAssAttr = visitGetattr = _visitAttr

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)

    def visitAssList(self, node):
        expr = self.expr
        self.expr = make_instance()
        for n in node.nodes:
            self.dispatch(n)
        self.expr = expr

    visitAssTuple = visitAssList

    def visitAssName(self, node):
        if self.expr:
            if isinstance(self.expr, Attr):
                expr = self.expr.get_value()
            elif isinstance(self.expr, TypedInstance):
                expr = self.expr
            else:
                return
        else:
            return

        value = node._values and node._values.get(node.name) or None

        # Need to replace any uncertain value with a concrete value.

        if value and isinstance(value, Instance) and not isinstance(value, TypedInstance):
            node._values[node.name] = expr

    def visitCallFunc(self, node):

        "Identify any concrete types involved with instantiation."

        for n in node.getChildNodes():
            self.dispatch(n)

        # Determine whether the target of the invocation refers to a class.

        attr = node.node._attr

        if attr:
            value = attr.get_value()
            if value and isinstance(value, Class):
                return TypedInstance(value)

# Convenience functions.

def deduce(program):
    for module in program.get_importer().get_modules():
        DeducedSource(module, program).deduce()

# vim: tabstop=4 expandtab shiftwidth=4
